# -*- coding: utf-8 -*-
"""minipippy test module."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals, with_statement)

if __name__ == "__main__":

    print("Mini Pippy - Super Awesome App")
