# -*- coding: utf-8 -*-
"""pyppyn setup script."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals, with_statement)

from setuptools import setup

if __name__ == "__main__":

    setup(
        name="minipippy",
        description="Not a real package, just for testing Pyppyn.",
        package_dir={'': 'src'}
    )
